//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#ifndef __KDRIVE_EXPRESS_ACCESS_H__
#define __KDRIVE_EXPRESS_ACCESS_H__

#include "kdrive_express_config.h"
#include "kdrive_express_defs.h"

/*!
	\file kdrive_express_access.h
	\brief kdriveExpress Access API
	\author Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
	\defgroup kdrive_express_access kdriveExpress Access API
*/

#define KDRIVE_MAX_TUNNEL_INFO_ITEMS		(63)     /*!< Maximum of tunneling info elements in ip_tunn_slot_info_t */

#ifdef __cplusplus
extern "C" {
#endif

/*******************************
** Init
********************************/

/*!
	Creates an internal AccessPort interface
	This should be the first function called when
	working with the access library. It creates a
	unique access port descriptor which will be passed to all
	kdrive access functions. When finished, call kdrive_ap_release.
	If create fails the function returns KDRIVE_INVALID_DESCRIPTOR
	\return access port descriptor if successful
	\ingroup kdrive_express_access
*/
kdriveExpress_API int32_t kdrive_ap_create(void);

/*!
	Releases the AccessPort interface.
	This should be the last function called when
	working with the access library.
	Returns true if successfully released
	false otherwise (i.e. the descriptor wasn't found).
	\param [in] ap the access port descriptor
	\ingroup kdrive_express_access
*/
kdriveExpress_API bool_t kdrive_ap_release(int32_t ap);

/*******************************
** Keyring
********************************/

/*!
	Sets the used knx keyring project. Use 0 when you want remove it
	\note Only available when KDRIVE_EXPRESS_SECURITY_ENABLED is 1
	\param [in] ap the access port descriptor
	\param [in] keyring the keyring project descriptor
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_keyring(int32_t ap, int32_t keyring);

/*******************************
** USB
********************************/

/*!
	Scan for all KNX USB Interface devices
	\param [in] ap the access port descriptor
	\return count of successfully enumerated interface devices
	\ingroup kdrive_express_access
*/
kdriveExpress_API uint32_t kdrive_ap_enum_usb(int32_t ap);

/*!
	Structure holding usb device information
	\see kdrive_ap_enum_usb_ex and kdrive_ap_only_enum_usb
	\ingroup kdrive_express_access
*/
typedef struct usb_dev_t
{
	uint16_t ind_addr; /*!< KNX individual address */
	uint16_t media_tytes; /*!< knx media types (bit field, encoding is DPT_Media)*/
	uint32_t internal_usb_index; /*!< numbered in the order that the hid driver enumerates the devices*/
	uint32_t usb_vendor_id; /*!< usb device vendor id*/
	uint32_t usb_product_id; /*!< usb device product id*/

} usb_dev_t;

/*!
	Scan for all KNX USB Interface devices

	Note: This function also opens and closes a connection
	to usb interface to read some info e.g. media type, individual address etc.
	It also list only not already opened usb interfaces.
	If you do not need this additional informations you could use kdrive_ap_only_enum_usb.

	The items array must exist (should be pre-allocated by the user).
	The max size of the array must be passed in as items_length.
	items_length will be updated to correspond to the length of the
	found devices. If the array is to small than only the possible values will returned.
	\param [out] items the array of usb device struct items with the found usb devices
	\param [in,out] items_length the data array capacity (in) and actual length of the data array (out)
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_enum_usb_ex(usb_dev_t items[], uint32_t* items_length);

/*!
	Scan for all KNX USB Interface devices

	This function does not open a connection to the usb interface so also already opened
	devices are seen but some information like media type and individual address is not available (set to 0).
	When you need this information see kdrive_ap_enum_usb_ex.

	Note: usb_dev_t.ind_addr and usb_dev_t.media_tytes are not read so the are not valid!

	The items array must exist (should be pre-allocated by the user).
	The max size of the array must be passed in as items_length.
	items_length will be updated to correspond to the length of the
	found devices. If the array is to small than only the possible values will returned.
	\param [out] items the array of usb device struct items with the found usb devices
	\param [in,out] items_length the data array capacity (in) and actual length of the data array (out)
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_only_enum_usb(usb_dev_t items[], uint32_t* items_length);

/*!
	Opens a connection to a KNX USB Interface device
	iface_index should be in the range 0..iface_count-1
	where iface_count is > 0

	Interfaces which supports this protocol:
	* all KNX conform USB interfaces
	* e.g. Weinzierl KNX USB Interface 310/311/320/330

	\sa kdriveEnumerateUSB
	\param [in] ap the access port descriptor
	\param [in] iface_index the interface device enumeration index
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_open_usb(int32_t ap, uint32_t iface_index);

/*******************************
** KNXnet/IP Tunneling
********************************/

/*!
	Structure holding ip tunnling slot information
	\see ip_tunn_dev_t
	\ingroup kdrive_express_access
*/
typedef struct ip_tunn_slot_info_t
{
	uint16_t address; /*!< the individual address of the tunnel */
	uint16_t status;  /*!< 2 bytes of status bits, currently only the last 3 are used, they are Usable(third bit), Authorized(second bit) and Free(last bit)
						   (KDRIVE_TUNNEL_SLOT_STATUS_FREE, KDRIVE_TUNNEL_SLOT_STATUS_AUTH, KDRIVE_TUNNEL_SLOT_STATUS_USABLE) */

} ip_tunn_slot_info_t;


/*!
	Structure holding ip tunneling device information
	\see kdrive_ap_enum_ip_tunn
	\ingroup kdrive_express_access
*/
typedef struct ip_tunn_dev_t
{
	// ip address and network interface
	char ip_address[KDRIVE_MAX_IP_ADDRESS_LEN]; /*!< IP address in the form "192.168.1.47" for example. Null-terminated string*/
	char iface_address[KDRIVE_MAX_IP_ADDRESS_LEN]; /*!< interface adaptor IP address. Null-terminated string*/

	// device information
	char dev_name[KDRIVE_MAX_DEVICE_NAME_LEN]; /*!< device friendly name. Null-terminated string in utf8 encoding*/
	uint16_t ind_addr; /*!< KNX individual address */
	uint8_t mac_address[KDRIVE_MAC_LEN]; /*!< mac address */
	uint8_t serial_number[KDRIVE_SN_LEN]; /*!< serial number */
	bool_t prog_mode_enabled; /*!< state of the programming mode 0 = off, 1 = on */

	// security information
	bool_t is_security_supported; /*!< supports the interface security (but it is possible that it is not configured) 0 = false, 1 = true */
	bool_t is_tunneling_secure; /*!< is tunneling service secure 0 = false, 1 = true */
	bool_t is_management_secure; /*!< is device management service secure 0 = false, 1 = true */

	// tunnling slot information
	// optional (When the device not support it then slot_infos_count is 0)
	uint16_t slot_infos_count; /*!< actual number of tunnling slot information items */
	ip_tunn_slot_info_t slot_infos[KDRIVE_MAX_TUNNEL_INFO_ITEMS]; /*!< tunnling slot information items */

} ip_tunn_dev_t;

/*!
	Scans for all KNX IP Tunneling Interface devices
	The items array must exist (should be pre-allocated by the user).
	The max size of the array must be passed in as items_length.
	items_length will be updated to correspond to the length of the
	found devices. If the array is to small than only the possible values will returned.
	\param [in] ap the access port descriptor
	\param [out] items the array of tunneling device struct items with the found ip tunning devices
	\param [in,out] items_length the data array capacity (in) and actual length of the data array (out)
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_enum_ip_tunn(int32_t ap,
        ip_tunn_dev_t items[],
        uint32_t* items_length);


/*!
	Gets the information for one specific tunneling device
	\param [in] ap the access port descriptor
	\param [in] ip_address the IP Address
	\param [in] port (optional)  the server port. When the port is 0 then the default port for KNXnet/IP (3671) is used
	\param [in] iface_address (optional) The Interface Adaptor IP Address <br />
	            If iface_address = 0 (null pointer) then the default network interface is used
	\param [out] device_info tunneling device struct item of the IP device
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_read_ip_tunn_device_info(int32_t ap,
        const char* ip_address,
        uint16_t port,
        const char* iface_address,
        ip_tunn_dev_t* device_info);

/*!
	Opens a connection to a KNX IP Tunneling Interface device
	on a specific network interface or default network interface

	Interfaces which supports this protocol:
	* all KNX conform KNXnet/IP Tunneling interfaces
	* e.g. Weinzierl KNX IP Interface 730/731/732
	* e.g. Weinzierl KNX IP Interface 740 Wireless
	* e.g. Weinzierl KNX IP Router 750/751/752
	* e.g. Weinzierl KNX IP LineMaster 760/762
	* e.g. Weinzierl KNX IP BAOS 770/771/772/773/774/777

	When a secure KNX IP Tunneling device is used then
	user id, password and authentication code (optional)
	must set before open.
	\sa kdrive_ap_set_ip_tunn_credentials
	\sa kdrive_ap_set_authentication_code

	or via keyring file
	\sa kdrive_ap_set_keyring
	\sa kdrive_ap_set_ip_tunn_user_id


	For use a given tunneling individual address set it with \sa kdrive_ap_set_ip_tunn_requested_ia
	before open.

	\sa kdrive_ap_open_ip_tunn_nat
	\param [in] ap the access port descriptor
	\param [in] ip_address the IP Address
	\param [in] port (optional)  the server port. When the port is 0 then the default port for KNXnet/IP (3671) is used
	\param [in] iface_address (optional) The Interface Adaptor IP Address <br />
	            If iface_address = 0 (null pointer) then the default network interface is used
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_open_ip_tunn(int32_t ap, const char* ip_address, uint16_t port, const char* iface_address);

/*!
	Opens a connection to a KNX IP Tunneling Interface device
	on a specific network interface in Network Address Translation (NAT) mode.
	\sa kdrive_ap_open_ip_tunn
	\param [in] ap the access port descriptor
	\param [in] ip_address the IP Address
	\param [in] port (optional)  the server port. When the port is 0 then the default port for KNXnet/IP (3671) is used
	\param [in] iface_address (optional) The Interface Adaptor IP Address <br />
	            If iface_address = 0 (null pointer) then the default network interface is used
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_open_ip_tunn_nat(int32_t ap, const char* ip_address, uint16_t port, const char* iface_address);

/*!
	\deprecated Use kdrive_ap_open_ip_tunn instead. Call it with iface_address=0.

	Opens a connection to a KNX IP Tunneling Interface device
	on default network interface

	\sa kdrive_ap_open_ip_tunn
	\sa kdrive_ap_open_ip_tunn_nat
	\param [in] ap the access port descriptor
	\param [in] ip_address The IP Address and (optional) the port which is separated by a colon.
			When the port is not set then the default port for KNXnet/IP (3671) is used
			e.g.
			"192.168.1.47"
			"192.168.1.47:3671"
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_open_ip(int32_t ap, const char* ip_address);

/*!
	\deprecated Was renamed to kdrive_ap_open_ip_tunn

	Opens a connection to a KNX IP Tunneling Interface device
	on a specific network interface or default network interface

	\sa kdrive_ap_open_ip_tunn
	\param [in] ap the access port descriptor
	\param [in] ip_address The IP Address and (optional) the port which is separated by a colon.
				When the port is not set then the default port for KNXnet/IP (3671) is used
				e.g.
				"192.168.1.47"
				"192.168.1.47:3671"
	\param [in] iface_address (optional) The Interface Adaptor IP Address <br />
	            If iface_address = 0 (null pointer) then the default network interface is used
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_open_ip_ex(int32_t ap, const char* ip_address, const char* iface_address);

/*!
	\deprecated Was renamed to kdrive_ap_open_ip_tunn_nat

	Opens a connection to a KNX IP Tunneling Interface device
	on a specific network interface in Network Address Translation (NAT) mode.
	\sa kdrive_ap_open_ip_tunn_nat
	\param [in] ap the access port descriptor
	\param [in] ip_address The IP Address and (optional) the port which is separated by a colon.
				When the port is not set then the default port for KNXnet/IP (3671) is used
				e.g.
				"192.168.1.47"
				"192.168.1.47:3671"
	\param [in] iface_address (optional) The Interface Adaptor IP Address <br />
	            If iface_address = 0 (null pointer) then the default network interface is used
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_open_ip_nat(int32_t ap, const char* ip_address, const char* iface_address);

/*!
	Only has works for knxnet ip tunneling ports, returns if security is enabled or not
	\param [in] ap the (tunneling) access port descriptor
	\param [out] secured is security used for tunning port
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_is_ip_tunn_secure(int32_t ap, bool_t* secured);

/*!
	Sets the optional authentication code for the KNXnet/IP tunneling connection.
	This function is used only for secure KNXnet/IP devices.
	Call this before open the connection. The default value is an empty string.
	\note This function is only for closed tunneling access ports
	\param [in] ap the access port descriptor
	\param [in] authentication_code the authentication code of the local device, an empty string means that no validation of the auth code will be done
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_ip_tunn_auth_code(int32_t ap, const char* authentication_code);

/*!
	Sets the user id and the security password for the KNXnet/IP tunneling connection.
	This function is used for secure KNXnet/IP devices.
	Call this before opening the connection.
	\note This function is only for closed tunneling access ports
	\param [in] ap the (tunneling) access port descriptor
	\param [in] user_id the security user id of the local device
	\param [in] password the security password of the local device
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_ip_tunn_credentials(int32_t ap, uint8_t user_id, const char* password);

/*!
	Sets the security user id for the KNXnet/IP tunneling connection.
	This function is used for secure KNXnet/IP devices.
	Call this before opening the connection. The default value is 1.
	\note This function is only for closed tunneling access ports
	\param [in] ap the access port descriptor
	\param [in] user_id the security user id of the local device
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_ip_tunn_user_id(int32_t ap, uint8_t user_id);

/*!
	Gets the security user id for the KNXnet/IP tunneling connection.
	\note This function is only for tunneling access ports
	\param [in] ap the access port descriptor
	\param [out] user_id the security user id of the local device
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_ip_tunn_user_id(int32_t ap, uint8_t* user_id);

/*!
	Sets the (optional) requested individual address
	This function is used for KNXnet/IP v2 devices. The default value is 0 ( = considered any).
	Call this before opening the connection.
	\note This function is only for closed tunneling access ports
	\param [in] ap the (tunneling) access port descriptor
	\param [in] address the individual address of the tunnel to open 0.0.0 is considered any
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_ip_tunn_requested_ia(int32_t ap, uint16_t address);

/*******************************
** KNXnet/IP Routing
********************************/

/*!
	Structure holding ip routing information
	\see kdrive_ap_enum_ip_rout
	\ingroup kdrive_express_access
*/
typedef struct ip_rout_info_t
{
	char multicast_address[KDRIVE_MAX_IP_ADDRESS_LEN]; /*!< Multicast IP address in the form "224.0.23.12" for example. Null-terminated string*/
	char iface_address[KDRIVE_MAX_IP_ADDRESS_LEN]; /*!< interface adaptor IP address. Null-terminated string*/
	uint32_t number_of_devices; /*!< Number of found KNXnet/IP Routing devices with the same multicast address and network interface */

} ip_rout_info_t;

/*!
	Scans for all KNX IP Routing devices
	The items array must exist (should be pre-allocated by the user).
	The max size of the array must be passed in as items_length.
	items_length will be updated to correspond to the length of the
	found devices. If the array is to small than only the possible values will returned.
	\note Only available in the Commercial Access kdriveExpress SDK
	\param [in] ap the access port descriptor
	\param [out] items the array of routing information struct items with the found ip routing devices
	\param [in,out] items_length the data array capacity (in) and actual length of the data array (out)
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_enum_ip_rout(int32_t ap,
        ip_rout_info_t items[],
        uint32_t* items_length);

/*!
	Creates an KNXnet/IP Routing access port

	When is_secure is true then backbone key must set before open.
	\sa kdrive_ap_set_ip_rout_backbone_key

	or via keyring file
	\sa kdrive_ap_set_keyring

	\note Only available in the Commercial Access kdriveExpress SDK
	\param [in] ap the access port descriptor
	\param [in] multicast_address (optional) The Multicast IP Address
				When the multicast_address = 0 (null pointer) then the default KNXnet/IP multicast address is used (224.0.23.12)
	\param [in] iface_address (optional) The Interface Adaptor IP Address <br />
	            If iface_address = 0 (null pointer) then the default network interface is used
	\param [in] secured with (1) or without encrypted comunication (0)
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_open_ip_rout(int32_t ap, const char* multicast_address, const char* iface_address, bool_t secured);

/*!
	Only has works for knxnet ip routing ports, returns if security is enabled or not
	\note Only available in the Commercial Access kdriveExpress SDK
	\param [in] ap the (routing) access port descriptor
	\param [out] secured is security enabled or disabled for routing port
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_is_ip_rout_secure(int32_t ap, bool_t* secured);

/*!
	Sets the backbone key for the KNXnet/IP routing connection.
	\note Only available in the Commercial Access kdriveExpress SDK
	\param [in] ap the access port descriptor
	\param [in] key the backbone key of the local device as hex string
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_ip_rout_backbone_key(int32_t ap, const char* key);

/*******************************
** Serial FT1.2
********************************/

/*!
	Opens a connection to a KNX FT1.2 serial interface.

	Interfaces which supports this protocol:
	* all KNX conform FT1.2 serial interfaces
	* e.g. Weinzierl KNX Serial BAOS 870
	* e.g. Weinzierl KNX BAOS Module 820/822/830/838

	\param [in] ap the access port descriptor
	\param [in] serial_device is the device where the interface is attached, i.e. COM1 or /dev/ttyS0 etc
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_open_serial_ft12(int32_t ap, const char* serial_device);

/*******************************
** TinySerial
********************************/

/*!
	Opens a connection to the KNX Tiny Serial Interface 810.
	This is a modified TP-UART Protocol without real-time requirements to the host

	Interfaces which supports this protocol:
	* Weinzierl KNX Tiny Serial Interface 810

	\param [in] ap the access port descriptor
	\param [in] serial_device is the device where the interface is attached, i.e. COM1 or /dev/ttyS0 etc
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_open_tiny_serial(int32_t ap, const char* serial_device);

/*******************************
** Connection Management
********************************/

/*!
	Closes the access port
	If the access port is not open nothing happens
	\param [in] ap the access port descriptor
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_close(int32_t ap);

/*!
	Returns whether the access port is open
	\param [in] ap the access port descriptor
	\return 1 if the access port is open, 0 otherwise
	\ingroup kdrive_express_access
*/
kdriveExpress_API bool_t kdrive_ap_is_open(int32_t ap);

/*******************************
** Send KNX Telegrams
********************************/

/*!
	Sends a KNX Telegram on the Access Port
	\param [in] ap the access port descriptor
	\param [in] telegram is the EMIC encoded KNX Telegram
	\param [in] telegram_len indicates the length of the telegram
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_send(int32_t ap, const uint8_t telegram[], uint32_t telegram_len);

/*!
	Sends a GroupValue_Write Telegram
	The length is specified in bits to enable values less than one byte
	to be sent (i.e. 1 bit boolean) etc.
	\param [in] ap the access port descriptor
	\param [in] address is the Group Address (destination)
	\param [in] value is the datapoint value to be sent
	\param [in] bits respresents the length of the datapoint value, in bits
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_group_write(int32_t ap, uint16_t address, const uint8_t* value, uint32_t bits);

/*!
	Sends a GroupValue_Read Telegram on the Access Port
	This function does not wait for a GroupValue_Response telegram
	\param [in] ap the access port descriptor
	\param [in] address is the Group Address (destination)
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_group_read(int32_t ap, uint16_t address);

/*!
	Sends a GroupValue_Response Telegram
	\sa kdriveGroupValueWrite
	\param [in] ap the access port descriptor
	\param [in] address is the Group Address (destination)
	\param [in] value is the datapoint value to be sent
	\param [in] bits respresents the length of the datapoint value, in bits
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_group_response(int32_t ap, uint16_t address, const uint8_t* value, uint32_t bits);

/*!
	Sends a GroupValue_Read Telegram on the Access Port
	and waits for a GroupValue_Response indication telegram
	(with a matching Group Address).
	If the response telegram has been received you can use
	kdrive_ap_get_group_data to get the response data.
	This function waits until either a telegram is received
	or the timeout elapses. If the timeout elapses the
	number of bytes returned is 0. On receipt of a matching
	GroupValue_Response telegram indication the corresponding
	telegram length is returned.
	\param [in] ap the access port descriptor
	\param [in] address is the Group Address (destination)
	\param [out] telegram buffer to hold the inbound telegram
	\param [in] telegram_len  As an input this holds the size of the telegram buffer
	\param [in] timeout The time period to wait in milliseconds for the response to arrive
	\ingroup kdrive_express_access
*/
kdriveExpress_API uint32_t kdrive_ap_read_group_object(int32_t ap, uint16_t address,
        uint8_t telegram[], uint32_t telegram_len, uint32_t timeout);

/*******************************
** Receive KNX Telegrams
********************************/

/*!
	Enables/Disables the receive queue for the access port.
	\param [in] ap the access port descriptor
	\param [in] enabled true enables the receive queue false disables the queue
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_enable_queue(int32_t ap, bool_t enabled);

/*!
	Waits for an inbound KNX telegram and (if received)
	copies the data into the telegram buffer.
	This function waits until either a telegram is received
	or the timeout elapses. If the timeout elapses the
	number of bytes returned is 0. On receipt of a telegram the
	corresponding telegram length is returned.
	\param [in] ap the access port descriptor
	\param [out] telegram buffer to hold the inbound telegram
	\param [in,out] telegram_len  As an input this holds the size of the telegram buffer
                	and as output the length of the received telegram
	\param [in] timeout The time period to wait in milliseconds for a telegram to arrive
	\ingroup kdrive_express_access
*/
kdriveExpress_API uint32_t kdrive_ap_receive(int32_t ap,
        uint8_t telegram[], uint32_t telegram_len, uint32_t timeout);

/*******************************
** KNX Telegram Callback
********************************/

/*!
	Callback function type (pointer to function definition)
	\sa kdrive_ap_register_telegram_callback
	\param [in] telegram holds the received telegram, the buffer is temporary
	\param [in] telegram_len indicates the length of the telegram
	\param [in] user_data pointer to user data or NULL
	\ingroup kdrive_express_access
*/
typedef void (*kdrive_ap_telegram_callback)(const uint8_t telegram[], uint32_t telegram_len, void* user_data);

/*!
	Registers a callback function.
	This callback function will be called when a telegram is received
	by the Access Port. A notification thread is used internally so
	this callback will be in the context of the notification thread
	(and not the main thread). That is, care should be taken when calling
	out from the callback.
	This function generates a unique key to represent the callback. This key
	can be used to remove the callback at a later state.
	\sa kdrive_ap_remove_telegram_callback
	\param [in] ap the access port descriptor
	\param [in] c the telegram callback pointer to function
	\param [in] user_data (optional) pointer to user data
	\param [out] key is set as a unique internal handle representing the callback function
	and can be used to remove the callback
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_register_telegram_callback(int32_t ap,
        kdrive_ap_telegram_callback c, void* user_data, uint32_t* key);

/*!
	Registers a callback function.
	This callback function will be called when a secure telegram is received
	by the Access Port. A notification thread is used internally so
	this callback will be in the context of the notification thread
	(and not the main thread). That is, care should be taken when calling
	out from the callback.
	This function generates a unique key to represent the callback. This key
	can be used to remove the callback at a later state.
	\sa kdrive_ap_remove_telegram_callback
	\param [in] ap the access port descriptor
	\param [in] c the telegram callback pointer to function
	\param [in] user_data (optional) pointer to user data
	\param [out] key is set as a unique internal handle representing the callback function
	and can be used to remove the callback
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_register_sec_telegram_callback(int32_t ap,
        kdrive_ap_telegram_callback c, void* user_data, uint32_t* key);

/*!
	Removes (de-registers) a telegram callback
	\param [in] ap the access port descriptor
	\param [in] key is the key returned from kdrive_ap_register_telegram_callback
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_remove_telegram_callback(int32_t ap, uint32_t key);

/*******************************
** Telegram Attributes
********************************/

/*!
	Extracts the Destination Address from the Telegram
	\param [in] telegram the KNX telegram
	\param [in] telegram_len the length of the KNX telegram
	\param [out] address the destination address
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_dest(const uint8_t telegram[],
        uint32_t telegram_len, uint16_t* address);

/*!
	Extracts the Source Address from the Telegram
	\param [in] telegram the KNX telegram
	\param [in] telegram_len the length of the KNX telegram
	\param [out] address the source address
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_src(const uint8_t telegram[],
        uint32_t telegram_len, uint16_t* address);

/*!
	Extracts the APCI from the Telegram
	\param [in] telegram the the KNX telegram
	\param [in] telegram_len the length of the KNX telegram
	\param [out] apci APCI
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_apci(const uint8_t telegram[],
        uint32_t telegram_len, uint16_t* apci);

/*!
	Extracts the Message Code from the Telegram
	\see KDRIVE_CEMI_L_DATA_REQ
	\see KDRIVE_CEMI_L_DATA_CON
	\see KDRIVE_CEMI_L_DATA_IND
	\param [in] telegram the KNX telegram
	\param [in] telegram_len the length of the KNX telegram
	\param [out] message_code the Message Code
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_message_code(const uint8_t telegram[],
        uint32_t telegram_len, uint8_t* message_code);

/*!
	Extracts the Serial Number from the Additional Info (if available)
	This is extracted from the Additional Info field
	i.e. for KNX-RF telegrams
	\param [in] telegram the KNX telegram
	\param [in] telegram_len the length of the KNX telegram
	\param [out] sn the Serial Number
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_serial_number(const uint8_t telegram[],
        uint32_t telegram_len, uint8_t sn[KDRIVE_SN_LEN]);

/*******************************
** Group Values
********************************/

/*!
	Determines if the specified telegram is a GroupValue telegram
	This is, one of GroupValue_Read, GroupValue_Write or GroupValue_Response.
	\param [in] telegram the KNX telegram
	\param [in] telegram_len the length of the KNX telegram
	\return true if a Group Telegram false otherwise
	\ingroup kdrive_express_access
*/
kdriveExpress_API bool_t kdrive_ap_is_group(const uint8_t telegram[], uint32_t telegram_len);

/*!
	Determines if the specified telegram is a GroupValue_Write telegram
	\param [in] telegram the KNX telegram
	\param [in] telegram_len the length of the KNX telegram
	\return true if a GroupValue_Write telegram false otherwise
	\ingroup kdrive_express_access
*/
kdriveExpress_API bool_t kdrive_ap_is_group_write(const uint8_t telegram[], uint32_t telegram_len);

/*!
	Determines if the specified telegram is a GroupValue_Read telegram
	\param [in] telegram the KNX telegram
	\param [in] telegram_len the length of the KNX telegram
	\return true if a GroupValue_Read telegram false otherwise
	\ingroup kdrive_express_access
*/
kdriveExpress_API bool_t kdrive_ap_is_group_read(const uint8_t telegram[], uint32_t telegram_len);

/*!
	Determines if the specified telegram is a GroupValue_Response telegram
	\param [in] telegram the KNX telegram
	\param [in] telegram_len the length of the KNX telegram
	\return true if a GroupValue_Response telegram false otherwise
	\ingroup kdrive_express_access
*/
kdriveExpress_API bool_t kdrive_ap_is_group_response(const uint8_t telegram[], uint32_t telegram_len);

/*!
	Extracts the Group Value data from the KNX telegram.
	If the telegram is not a GroupValue_Write or GroupValue_Response telegram
	an Invalid Telegram error is returned.
	The data buffer should be large enough to hold the data from the telegram.
	This length is checked against the length held in the telegram and if insufficient
	an Access Port Error is returned (Buffer Overflow).
	\param [in] telegram the KNX Telegram (Group Value Telegram)
	\param [in] telegram_len the length of the telegram
	\param [out] data The data buffer
	\param [in,out] data_len the data buffer capacity (in) and actual length of the data buffer (out)
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_group_data(const uint8_t telegram[],
        uint32_t telegram_len, uint8_t* data, uint32_t* data_len);

/*!
	Determines if the group value size is compressed (6 bits or less).
	If the telegram is not a GroupValue_Write or GroupValue_Response telegram
	an Invalid Telegram error is returned.
	\param [in] telegram the KNX Telegram (Group Value Telegram)
	\param [in] telegram_len the length of the telegram
	\param [out] is_commpressed 1 : compressed (6 bits or less) or 0 : not compressed (>= 1 byte)
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_is_compressed_group_data(const uint8_t telegram[],
        uint32_t telegram_len, bool_t* is_commpressed);

/*******************************
** Data Security
********************************/

enum data_sec_t
{
	Off  = 0x00, /*!< KNX individual address */
	On   = 0x01, /*!< knx media types (bit field, encoding is DPT_Media)*/
	Auto = 0x03  /*!< numbered in the order that the hid driver enumerates the devices*/
};

/*!
	Set the Data Security of the access point.
	It can be set to:
	 * KDRIVE_ACCESS_DATA_SEC_OFF  = 0x00
	 * KDRIVE_ACCESS_DATA_SEC_ON   = 0x01
	 * KDRIVE_ACCESS_DATA_SEC_AUTO = 0xFF
	\note Only available when KDRIVE_EXPRESS_SECURITY_ENABLED is 1
	\param [in] ap the access port descriptor
	\param [in] dataSec Data Security usage to set
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_data_sec(int32_t ap, enum data_sec_t dataSec);

/*!
	Get the Data Security usage of the access point.
	It can be:
	 * KDRIVE_ACCESS_DATA_SEC_OFF  = 0x00
	 * KDRIVE_ACCESS_DATA_SEC_ON   = 0x01
	 * KDRIVE_ACCESS_DATA_SEC_AUTO = 0xFF
	\note Only available when KDRIVE_EXPRESS_SECURITY_ENABLED is 1
	\param [in] ap the access port descriptor
	\param [out] dataSec current Data Security usage
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_data_sec(int32_t ap, enum data_sec_t* dataSec);

/*!
	Set the own sending sequence number.
	\note Only available when KDRIVE_EXPRESS_SECURITY_ENABLED is 1
	\param [in] ap the access port descriptor
	\param [in] seq_num own sending sequence number
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_data_sec_own_seq_num(int32_t ap, const uint8_t seq_num[6]);

/*!
	Set the sequence number of a device.
	\note Only available when KDRIVE_EXPRESS_SECURITY_ENABLED is 1
	\param [in] ap the access port descriptor
	\param [in] address individual address of the device
	\param [in] seq_num the sequence number of the device
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_data_sec_device_seq_num(int32_t ap, uint16_t address, const uint8_t seq_num[6]);

/*!
	Check if the telegram is a data secure one.
	\param [in] telegram the KNX Telegram
	\param [in] telegram_len the length of the telegram
	\ingroup kdrive_express_access
*/
kdriveExpress_API bool_t kdrive_ap_is_data_sec(const uint8_t telegram[], uint32_t telegram_len);

/*******************************
** Access Port Properties
********************************/

/*!
	Gets the Layer of the Local Device (KNX Interface Device)
	The default value is Link Layer.
	\param [in] ap the access port descriptor
	\param [out] layer the KNX layer of the local device
	\ingroup kdrive_express_access
	\see kdrive_ap_set_layer
*/
kdriveExpress_API error_t kdrive_ap_get_layer(int32_t ap, uint16_t* layer);

/*!
	Sets the Layer of the Local Device (KNX Interface Device)
	\note Only available in the Commercial Access kdriveExpress SDK
	\param [in] ap the access port descriptor
	\param [in] layer the KNX layer of the local device
	KDRIVE_LAYER_LINK, KDRIVE_LAYER_BUSMON or KDRIVE_LAYER_RAW
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_layer(int32_t ap, uint16_t layer);

/*!
	Gets the Individual Address of the Local Device (KNX Interface Device)
	\param [in] ap the access port descriptor
	\param [out] address the individual address of the local device
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_ind_addr(int32_t ap, uint16_t* address);

/*!
	Sets the Individual Address of the Local Device (KNX Interface Device)
	\param [in] ap the access port descriptor
	\param [in] address the individual address of the local device
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_ind_addr(int32_t ap, uint16_t address);

/*!
	Gets the Domain Address of the Local Device (KNX Interface Device)
	\param [in] ap the access port descriptor
	\param [out] da the domain address of the local device
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_rf_domain_addr(int32_t ap, uint8_t da[KDRIVE_DA_LEN]);

/*!
	Sets the Domain Address of the Local Device (KNX Interface Device)
	\param [in] ap the access port descriptor
	\param [in] da the new domain address of the local device
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_rf_domain_addr(int32_t ap, const uint8_t da[KDRIVE_DA_LEN]);

/*!
	Gets the Serial Number of the Local Device (KNX Interface Device)
	\param [in] ap the access port descriptor
	\param [out] sn the serial number of the local device
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_serial_no(int32_t ap, uint8_t sn[KDRIVE_SN_LEN]);

/*!
	Sets the Serial Number of the Local Device (KNX Interface Device)
	\param [in] ap the access port descriptor
	\param [in] sn the new serial number of the local device
	\ingroup kdrive_express_access
	\note Normally the serial number is write protected. Only some interfaces for testing are support writing.
*/
kdriveExpress_API error_t kdrive_ap_set_serial_no(int32_t ap, const uint8_t sn[KDRIVE_SN_LEN]);

/*!
	Gets the used Individual Address of the tunnel connection (KNX Interface Device).
	This address is the device address PID_KNX_INDIVIDUAL_ADDRESS or
	a additional individual address PID_ADDITIONAL_INDIVIDUAL_ADDRESSES (normal case).
	The interface communicates it at the connecting procedure.
	So if somebody changed the additional individual addresses than this value is not up to data!
	\note This function is only for tunneling access ports.
	\param [in] ap the access port descriptor of a tunneling access port
	\param [out] address the used individual address for the tunnel
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_tunnel_ind_addr(int32_t ap, uint16_t* address);

/*!
	Sets the used Individual Address of the tunnel connection (KNX Interface Device).
	Returns an error, if it can not changed because for the tunnel address is the device individual address used.
	\note This function is only for tunneling access ports
	\param [in] ap the access port descriptor of a tunneling access port
	\param [in] address the individual address of the tunnel
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_tunnel_ind_addr(int32_t ap, uint16_t address);

/*!
	Gets the Additional Individual Addresses of the Local Device (KNX Interface Device)
	\note This function is only for tunneling access ports
	\param [in] ap the access port descriptor of a tunneling access port
	\param [out] addresses the array of additional individual addresses
	\param [in,out] addresses_len the array capacity (in) and actual length of the array (out)
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_additional_ind_addr(int32_t ap, uint16_t addresses[], uint32_t* addresses_len);

/*!
	Sets the Additional Individual Addresses of the Local Device (KNX Interface Device)
	\note This function is only for tunneling access ports
	\param [in] ap the access port descriptor of a tunneling access port
	\param [out] addresses the array of additional individual addresses
	\param [in] addresses_len the array length
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_additional_ind_addr(int32_t ap, const uint16_t addresses[], uint16_t addresses_len);

/*!
	Gets the destination individual address filter setting.
	The default is 1 (filter is enabled).
	\note This function is only for KNXnet/IP Routing or Weinzierl cEMI FT1.2 interfaces.
	\param [in] ap the access port descriptor of a FT1.2 or routing access port
	\param [out] is_enabled 1 if filter is enabled or 0 otherwise
	\ingroup kdrive_express_access
	\see kdrive_ap_set_filter_dest_addr
*/
kdriveExpress_API error_t kdrive_ap_get_filter_dest_addr(int32_t ap, bool_t* is_enabled);

/*!
	Enable or disable filter destination individual address for L-Data.ind telegrams.
	If enabled (=1) than individual addressed telegrams for other destinations will not route.
	Only available for Weinzierl cEMI FT1.2 Interface.
	The other interfaces filter the individual address already self in the link layer.
	\note This function is only for KNXnet/IP Routing or Weinzierl cEMI FT1.2 interfaces.
	\note Only available in the Commercial Access kdriveExpress SDK
	\param [in] ap the access port descriptor of a FT1.2 or routing access port
	\param [in] enable enable (1) or disable (0) the filter
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_filter_dest_addr(int32_t ap, bool_t enable);

/*!
	Gets the rf domain adrress filter setting.
	The default is 1 (filter is enabled).
	\note This function is only for RF interfaces.
	\param [in] ap the access port descriptor
	\param [out] is_enabled 1 if filter is enabled or 0 otherwise
	\ingroup kdrive_express_access
	\see kdrive_ap_set_filter_rf_domain_addr
*/
kdriveExpress_API error_t kdrive_ap_get_filter_rf_domain_addr(int32_t ap, bool_t* is_enabled);

/*!
	Enable or disable filter rf domain address for L-Data.ind telegrams.
	If enabled (=1) than telegrams for other domains will not route.
	Only available, if medium of interface is RF. If it has a other medium nothing happens.
	\note This function is only for RF interfaces.
	\note Only available in the Commercial Access kdriveExpress SDK
	\param [in] ap the access port descriptor
	\param [in] enable enable (1) or disable (0) the filter
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_filter_rf_domain_addr(int32_t ap, bool_t enable);

/*!
	Reads the current bus state from the interface (use_cache == false) or
	gets the last status of the knx bus from the cache (use_cache == true).
	Normally the interface inform the client when the bus state changed.
	\note This function is only for KNX USB interfaces.
	\param [in] ap the access port descriptor
	\param [in] use_cache read from interface (false) or use last known status (true)
	\param [out] bus_state the current knx bus state
		(KDRIVE_KNX_BUS_STATE_UNKNOWN, KDRIVE_KNX_BUS_STATE_DISCONNECTED or KDRIVE_KNX_BUS_STATE_CONNECTED)
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_bus_state(int32_t ap, bool_t use_cache, uint32_t* bus_state);

/*!
	Gets the maximum supported apdu length of the Local Device (KNX Interface Device)
	\param [in] ap the access port descriptor
	\param [out] max_apdu_length the maximum apdu length from the interface
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_get_max_apdu_length(int32_t ap, uint32_t* max_apdu_length);

/*******************************
** Packet Trace
********************************/

/*!
	Starts the Packet Trace, for a specific access port.
	When no packet trace callback was set (the default) then it
	write Rx and Tx packets to the logger (level: information),
	otherwise it calls the callback function.
	precondition: the access port is open
	\param [in] ap the access port descriptor
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_packet_trace_connect(int32_t ap);

/*!
	Stops the Packet Trace
	\param [in] ap the access port descriptor
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_packet_trace_disconnect(int32_t ap);

/*!
	Packet Trace callback function type (pointer to function definition)
	\sa kdrive_ap_set_packet_trace_callback
	\param [in] telegram holds the received telegram, the buffer is temporary
	\param [in] telegram_len indicates the length of the telegram
	\param [in] direction the packet direction: KDRIVE_PACKET_DIR_TX (send) or KDRIVE_PACKET_DIR_RX (received)
	\param [in] user_data pointer to user data or NULL
	\ingroup kdrive_express_access
*/
typedef void(*kdrive_ap_packet_trace_callback)(const uint8_t telegram[], uint32_t telegram_len, int32_t direction, void* user_data);

/*!
	Sets the packet trace callback function, for a specific access port.
	This callback function will be called when a packet rx or tx by the Access Port.
	A notification thread is used internally so
	this callback will be in the context of the notification thread
	(and not the main thread). That is, care should be taken when calling
	out from the callback.
	There can only be one callback per access port.
	To remove the callback add an empty callback (0).
	\param [in] ap the access port descriptor
	\param [in] c the packet trace callback pointer to function
	\param [in] user_data (optional) pointer to user data
	\ingroup kdrive_express_access
*/
kdriveExpress_API error_t kdrive_ap_set_packet_trace_callback(int32_t ap, kdrive_ap_packet_trace_callback c, void* user_data);

#ifdef __cplusplus
}
#endif

#endif /* __KDRIVE_EXPRESS_ACCESS_H__ */
