//
// Copyright (c) 2002-2023 WEINZIERL ENGINEERING GmbH
// All rights reserved.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY,
// WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
//

#include <stdio.h>
#include <stdlib.h>
#include <kdrive_express.h>

/*******************************
** Main
********************************/

int main(int argc, char* argv[])
{
	uint16_t address = 0x901; /*!< The KNX Group Address (destination address) we use to send with */
	uint8_t value = 0; /*!< The value we send on the bus with the Group Value Write command */
	int32_t ap = 0; /*! the Access Port descriptor */
	int32_t keyring = 0; /*! the Keyring descriptor */
	const char* keyring_file = "data/keyring.knxkeys"; /*! the path to keyring project file */
	const char* keyring_password = "123456"; /*! the password for the keyring project file */
	uint32_t project_name_length = 128; /*! the maximal length of the keyring project name */
	char project_name[128]; /*! the keyring project name */
	uint32_t project_creator_length = 32; /*! the maximal length of the keyring project creator */
	char project_creator[32]; /*! the keyring project creator */

	/* Configure the logging level and console logger */
	kdrive_logger_set_level(KDRIVE_LOGGER_INFORMATION);
	kdrive_logger_console();

	/*
		We create a Access Port descriptor. This descriptor is then used for
		all calls to that specific access port.
	*/
	ap = kdrive_ap_create();

	/*
		We check that we were able to allocate a new descriptor
		This should always happen, unless a bad_alloc exception is internally thrown
		which means the memory couldn't be allocated.
	*/
	if (ap == KDRIVE_INVALID_DESCRIPTOR)
	{
		kdrive_logger(KDRIVE_LOGGER_FATAL, "Unable to create access port. This is a terminal failure");
		return -1;
	}

	/*
		Import the ETS keyring project
	*/
	if (kdrive_keyring_import_knxkeys(keyring_file, keyring_password, &keyring) != KDRIVE_ERROR_NONE)
	{
		kdrive_logger(KDRIVE_LOGGER_FATAL, "Unable to import the keyring file");
		return -1;
	}

	/*
		Trace some keyring project info
	*/
	kdrive_keyring_get_proj_name(keyring, project_name, &project_name_length);
	kdrive_logger_ex(KDRIVE_LOGGER_INFORMATION, "Project name %s", project_name);

	kdrive_keyring_get_proj_creator(keyring, project_creator, &project_creator_length);
	kdrive_logger_ex(KDRIVE_LOGGER_INFORMATION, "Created by: %s", project_creator);


	/*
		Use the keyring project for the tunneling port
		(The password, etc will set automatic from the keyring project)
	*/
	if (kdrive_ap_set_keyring(ap, keyring) != KDRIVE_ERROR_NONE)
	{
		kdrive_logger(KDRIVE_LOGGER_FATAL, "Unable to use the keyring project");
		return -1;
	}

	/* sets security user id (optional; default is 1)*/
	if (kdrive_ap_set_ip_tunn_user_id(ap, 2) != KDRIVE_ERROR_NONE)
	{
		kdrive_logger(KDRIVE_LOGGER_FATAL, "Unable to set the user id");
	}

	/*
		Open a Tunneling connection with a specific IP Interface,
		you will probably have to change the IP address
	*/
	if (kdrive_ap_open_ip_tunn(ap, "192.168.1.6", 0, 0) == KDRIVE_ERROR_NONE)
	{
		/*
			Connect the Packet Trace logging mechanism
			to see the Rx and Tx packets
		*/
		kdrive_ap_packet_trace_connect(ap);

		/* send a 1-Bit boolean GroupValueWrite telegram: on */
		value = 1;
		kdrive_ap_group_write(ap, address, &value, 1);

		/* send a 1-Bit boolean GroupValueWrite telegram: off */
		value = 0;
		kdrive_ap_group_write(ap, address, &value, 1);

		/* close the access port */
		kdrive_ap_close(ap);
	}

	/* releases the access port */
	kdrive_ap_release(ap);

	return 0;
}

/*******************************
** Private Functions
********************************/

/*
	Replace the following licence with your license key
*/
error_t set_license()
{
	char license[] =
	    "[Check]\n"
	    "Key=I1EkM2JvaE1HMEZtIlF5VnJuJn08KHRfV0AwTlEwRyk4NT55JFNvXnU6ZnpzMDI6QV42JWZ4"
	    "fHNed11aOld9WHpGXSEibnkrSDVsZU9WIkljeSQ3NWY7eT4rdEJ6XjhpNnw+OkNEIydTYWNF"
	    "RE15JVdQOno3XmtOJjZaRExwLSV8ZXxOYzlJITlhfFFdOHIjUW43SEshfSx8RWgzSipUZU5X"
	    "ZiJpSnIhRS9pJyQrOis/VVAoSHMkaV5kYGFfaFxhJj5peUtdfVBgMGh7K25bITF7OnhdeTQv"
	    "NiZlWCt4Nls3S31nLUU2LE1OXmhubGtxY2wqJHMxVDd0Tlg4OGlIUCE=\n"
	    "\n"
	    "[Customer]\n"
	    "City=84508 Burgkirchen / Alz\n"
	    "Company=WEINZIERL ENGINEERING GmbH\n"
	    "Country=Germany\n"
	    "Street=Achatz 3\n"
	    "\n"
	    "[LicenseData]\n"
	    "LicNum=kdrive-2012-05-14\n"
	    "LicVersion=71949c2a-be30-4fbe-9c96-d0a617ac6ab8\n"
	    "\n"
	    "[Limitations]"
	    "DateLimit=2012-12-31"
	    "\n"
	    "[ProgramFeatures]\n"
	    "kdriveAccess=0\n"
	    "kdriveServices=0\n";

	return kdrive_lic_set_license(license);
}

